#include <Windows.h>
#include <Shlwapi.h>
#include <MSTask.h>
#include <Lmcons.h>
#include "resource.h"
#include "Util.h"
#include "RecordingOption.h"
#include "ReserveList.h"


CReserveList::CReserveList()
    : m_head(NULL)
    , m_hThread(NULL)
{
    m_saveFileName[0] = 0;
    m_saveTaskName[0] = 0;
    m_pluginShortPath[0] = 0;
}


CReserveList::~CReserveList()
{
    Clear();
    if (m_hThread) {
        ::WaitForSingleObject(m_hThread, INFINITE);
        ::CloseHandle(m_hThread);
    }
}


void CReserveList::Clear()
{
    CBlockLock lock(&m_writeLock);

    while (m_head) {
        RESERVE *pRes = m_head;
        m_head = pRes->next;
        delete pRes;
    }
}


// strɂ͏ȂƂ1024vf̊mۂKv
void CReserveList::ToString(const RESERVE &res, LPTSTR str)
{
    TCHAR szStartTime[64];
    TCHAR szDuration[64];

    FileTimeToStr(&res.startTime, szStartTime);
    TimeSpanToStr(res.duration, szDuration);
    int len = ::wsprintf(str, TEXT("0x%04X\t0x%04X\t0x%04X\t0x%04X\t%s\t%s\t%s\t"),
                         res.networkID, res.transportStreamID, res.serviceID, res.eventID,
                         szStartTime, szDuration, res.eventName);
    res.recOption.ToString(str + len);
}


bool CReserveList::Insert(const RESERVE &in)
{
    RESERVE *pRes;
    RESERVE *prev = NULL;

    // \񂪂Ȃׂ
    for (pRes = m_head; pRes; pRes = pRes->next) {
        if (pRes->eventID == in.eventID &&
            pRes->networkID == in.networkID &&
            pRes->transportStreamID == in.transportStreamID &&
            pRes->serviceID == in.serviceID) break;
        prev = pRes;
    }

    CBlockLock lock(&m_writeLock);

    if (pRes) {
        // xXg؂藣
        if (!prev) m_head = pRes->next;
        else prev->next = pRes->next;
    }
    else {
        pRes = new RESERVE;
    }
    *pRes = in;

    // ̓`FbN
    ReplaceTokenDelimiters(pRes->eventName);
    ReplaceTokenDelimiters(pRes->recOption.saveDir);
    ReplaceTokenDelimiters(pRes->recOption.saveName);

    if (!m_head) {
        pRes->next = NULL;
        m_head = pRes;
    }
    else if (m_head->GetTrimmedStartTime() - pRes->GetTrimmedStartTime() > 0) {
        // \Xg̐擪ɑ}
        pRes->next = m_head;
        m_head = pRes;
    }
    else {
        // \Xgɑ}
        RESERVE *tail;
        for (tail = m_head; tail->next; tail = tail->next) {
            if (tail->next->GetTrimmedStartTime() - pRes->GetTrimmedStartTime() > 0) break;
        }
        pRes->next = tail->next;
        tail->next = pRes;
    }

    return true;
}


bool CReserveList::Insert(LPCTSTR str)
{
    RESERVE res;
    int i = 0;

    ::StrToIntEx(str, STIF_SUPPORT_HEX, &i);
    res.networkID = static_cast<WORD>(i);
    if (!NextToken(&str)) return false;

    ::StrToIntEx(str, STIF_SUPPORT_HEX, &i);
    res.transportStreamID = static_cast<WORD>(i);
    if (!NextToken(&str)) return false;

    ::StrToIntEx(str, STIF_SUPPORT_HEX, &i);
    res.serviceID = static_cast<WORD>(i);
    if (!NextToken(&str)) return false;

    ::StrToIntEx(str, STIF_SUPPORT_HEX, &i);
    res.eventID = static_cast<WORD>(i);
    if (!NextToken(&str)) return false;

    if (!StrToFileTime(str, &res.startTime)) return false;
    if (!NextToken(&str)) return false;

    if (!StrToTimeSpan(str, &res.duration)) return false;
    if (!NextToken(&str)) return false;

    GetToken(str, res.eventName, ARRAY_SIZE(res.eventName));
    if (!NextToken(&str)) return false;

    if (!res.recOption.FromString(str)) return false;

    return Insert(res);
}


bool CReserveList::Insert(HINSTANCE hInstance, HWND hWndParent, const RESERVE &in,
                          const RECORDING_OPTION &defaultRecOption, LPCTSTR serviceName, LPCTSTR captionSuffix)
{
    DIALOG_PARAMS prms = { in, &defaultRecOption, serviceName, captionSuffix };
    INT_PTR rv = ::DialogBoxParam(hInstance, MAKEINTRESOURCE(IDD_RESERVATION), hWndParent,
                                  DlgProc, reinterpret_cast<LPARAM>(&prms));

    return rv == IDOK ? Insert(prms.res) :
           rv == IDC_DELETE ? Delete(prms.res.networkID, prms.res.transportStreamID,
                                     prms.res.serviceID, prms.res.eventID) : false;
}


INT_PTR CALLBACK CReserveList::DlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg) {
    case WM_INITDIALOG:
        {
            DIALOG_PARAMS *pPrms = reinterpret_cast<DIALOG_PARAMS*>(lParam);
            RESERVE *pRes = &pPrms->res;
            ::SetWindowLongPtr(hDlg, GWLP_USERDATA, reinterpret_cast<LONG_PTR>(pRes));

            // LvV
            TCHAR cap[128];
            if (pPrms->captionSuffix && pPrms->captionSuffix[0] && ::GetWindowText(hDlg, cap, 32)) {
                ::lstrcpyn(cap + ::lstrlen(cap), pPrms->captionSuffix, 32);
                ::SetWindowText(hDlg, cap);
            }

            if (pPrms->serviceName && pPrms->serviceName[0]) {
                ::SetDlgItemText(hDlg, IDC_STATIC_SERVICE_NAME, pPrms->serviceName);
            }

            SYSTEMTIME sysTime;
            FILETIME time = pRes->GetTrimmedStartTime();
            ::FileTimeToSystemTime(&time, &sysTime);
            TCHAR text[512];
            ::wsprintf(text, TEXT("%dN%d%d(%s) %d%d%db  %d%db"),
                       sysTime.wYear, sysTime.wMonth, sysTime.wDay, GetDayOfWeekText(sysTime.wDayOfWeek),
                       sysTime.wHour, sysTime.wMinute, sysTime.wSecond,
                       pRes->GetTrimmedDuration() / 60, pRes->GetTrimmedDuration() % 60);
            ::SetDlgItemText(hDlg, IDC_STATIC_RES_TIME, text);

            ::SetDlgItemText(hDlg, IDC_EDIT_EVENT_NAME, pRes->eventName);
            ::SendDlgItemMessage(hDlg, IDC_EDIT_EVENT_NAME, EM_LIMITTEXT, ARRAY_SIZE(pRes->eventName) - 1, 0);

            return pRes->recOption.DlgProc(hDlg, uMsg, wParam, true, pPrms->pDefaultRecOption);
        }
    case WM_COMMAND:
        switch (LOWORD(wParam)) {
        case IDOK:
            {
                RESERVE *pRes = reinterpret_cast<RESERVE*>(::GetWindowLongPtr(hDlg, GWLP_USERDATA));
                if (!::GetDlgItemText(hDlg, IDC_EDIT_EVENT_NAME, pRes->eventName, ARRAY_SIZE(pRes->eventName)))
                    pRes->eventName[0] = 0;
                pRes->recOption.DlgProc(hDlg, uMsg, wParam, true);
            }
            // fall through!
        case IDC_DELETE:
        case IDCANCEL:
            ::EndDialog(hDlg, LOWORD(wParam));
            return TRUE;
        default:
            {
                RESERVE *pRes = reinterpret_cast<RESERVE*>(::GetWindowLongPtr(hDlg, GWLP_USERDATA));
                return pRes->recOption.DlgProc(hDlg, uMsg, wParam, true);
            }
        }
        break;
    }
    return FALSE;
}


bool CReserveList::Delete(DWORD networkID, DWORD transportStreamID, DWORD serviceID, DWORD eventID)
{
    RESERVE *tail;
    RESERVE *prev = NULL;

    for (tail = m_head; tail; tail = tail->next) {
        if (tail->eventID == eventID &&
            tail->networkID == networkID &&
            tail->transportStreamID == transportStreamID &&
            tail->serviceID == serviceID) break;
        prev = tail;
    }

    if (!tail) return false;

    CBlockLock lock(&m_writeLock);

    if (!prev) m_head = tail->next;
    else prev->next = tail->next;
    delete tail;
    return true;
}


const RESERVE *CReserveList::Get(DWORD networkID, DWORD transportStreamID, DWORD serviceID, DWORD eventID) const
{
    RESERVE *tail = m_head;
    for (; tail; tail = tail->next) {
        if (tail->eventID == eventID &&
            tail->networkID == networkID &&
            tail->transportStreamID == transportStreamID &&
            tail->serviceID == serviceID) break;
    }
    return tail;
}


const RESERVE *CReserveList::Get(int index) const
{
    if (index < 0) return NULL;

    RESERVE *tail = m_head;
    for (int i = 0; i < index && tail; i++, tail = tail->next);
    return tail;
}


bool CReserveList::Load()
{
    Clear();
    if (!::PathFileExists(m_saveFileName)) return true;

    LPTSTR text = NULL;
    for (int i = 0; i < 5; ++i) {
        if ((text = NewReadTextFileToEnd(m_saveFileName, FILE_SHARE_READ)) != NULL) break;
        ::Sleep(200);
    }
    if (!text) return false;

    // Inserť̂(債ĕς񂯂)tɓǂ
    LPCTSTR line = text + ::lstrlen(text);
    do {
        line = ::StrRChr(text, line, TEXT('\n'));
        if (!Insert(!line ? text : line + 1)) {
            DEBUG_OUT(TEXT("CReserveList::Load(): Insert Error\n"));
        }
    } while (line);

    delete [] text;
    return true;
}


bool CReserveList::Save() const
{
    HANDLE hFile = ::CreateFile(m_saveFileName, GENERIC_WRITE, 0, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile == INVALID_HANDLE_VALUE) return false;

    DWORD writtenBytes;
    WCHAR bom = L'\xFEFF';
    ::WriteFile(hFile, &bom, sizeof(bom), &writtenBytes, NULL);

    for (RESERVE *tail = m_head; tail; tail = tail->next) {
        TCHAR buf[1024 + 2];
        ToString(*tail, buf);
        ::lstrcat(buf, TEXT("\r\n"));
        ::WriteFile(hFile, buf, ::lstrlen(buf) * sizeof(TCHAR), &writtenBytes, NULL);
    }

    ::CloseHandle(hFile);
    return true;
}


#define GET_PRIORITY(x) ((x) % PRIORITY_MOD == PRIORITY_DEFAULT ? (x) + defaultRecOption.priority % PRIORITY_MOD : (x))
#define GET_START_MARGIN(x) ((x) < 0 ? defaultRecOption.startMargin : (x))


// ߂̗\擾
RESERVE *CReserveList::GetNearest(const RECORDING_OPTION &defaultRecOption, RESERVE **pPrev) const
{
    *pPrev = NULL;

    RESERVE *pMinRes = NULL;
    FILETIME minStart;
    minStart.dwLowDateTime = 0xFFFFFFFF;
    minStart.dwHighDateTime = 0x7FFFFFFF;

    // Jn}[W܂߂ĂƂ߂̗\T
    for (RESERVE *tail = m_head; tail; tail = tail->next) {
        FILETIME start = tail->GetTrimmedStartTime();
        start += -GET_START_MARGIN(tail->recOption.startMargin) * FILETIME_SECOND;

        // ̗\͗Dx̍̂I
        if (minStart - start > 0 || minStart - start == 0 &&
            GET_PRIORITY(pMinRes->recOption.priority) < GET_PRIORITY(tail->recOption.priority))
        {
            *pPrev = pMinRes;
            pMinRes = tail;
            minStart = start;
        }
        // Xg͊\[gȂ̂MARGIN_MAXbȏx\TKv͂Ȃ
        if (start - minStart > MARGIN_MAX * FILETIME_SECOND) break;
    }

    return pMinRes;
}


const RESERVE *CReserveList::GetNearest(const RECORDING_OPTION &defaultRecOption) const
{
    RESERVE *prev;
    return GetNearest(defaultRecOption, &prev);
}


// ߂̗\擾
// ErecOption̓ftHgKpς݂ɂȂ
// EstartTrimendTrim0ɕ␳(gς݂ɂȂ)
// EstartTimeduration͗\̏dɂĒꍇ
bool CReserveList::GetNearest(RESERVE *pRes, const RECORDING_OPTION &defaultRecOption, int readyOffset) const
{
    const RESERVE *pNearest = GetNearest(defaultRecOption);
    if (!pNearest || !pRes) return false;
    RESERVE &res = *pRes;
    res = *pNearest;
    res.recOption.ApplyDefault(defaultRecOption);

    // gς݂ɂ
    res.startTime = res.GetTrimmedStartTime();
    res.duration = res.GetTrimmedDuration();
    res.recOption.startTrim = 0;
    res.recOption.endTrim = 0;

    // I}[W͘^掞Ԃ𒴂ĕłĂ͂ȂȂ
    if (res.recOption.endMargin < -res.duration) res.recOption.endMargin = -res.duration;

    FILETIME resEnd = res.startTime;
    resEnd += (res.duration + res.recOption.endMargin) * FILETIME_SECOND;

    // Dx̍ʂ̗\񂪂Θ^I𑁂߂
    FILETIME resRealEnd = resEnd;
    for (const RESERVE *tail = m_head; tail; tail = tail->next) {
        FILETIME start = tail->GetTrimmedStartTime();
        // ^͂ɐ؂ւȂ̂ŁAreadyOffsetb̗]T
        start += -(GET_START_MARGIN(tail->recOption.startMargin) + readyOffset) * FILETIME_SECOND;

        if (res.recOption.priority < GET_PRIORITY(tail->recOption.priority) &&
            resRealEnd - start > 0) resRealEnd = start;

        if (start - resRealEnd > MARGIN_MAX * FILETIME_SECOND) break;
    }

    int diff = static_cast<int>((resEnd - resRealEnd) / FILETIME_SECOND);
    // I}[W
    if (diff <= res.recOption.endMargin) {
        res.recOption.endMargin -= diff;
    }
    else {
        // I}[ŴƂdiff͑
        diff -= res.recOption.endMargin;
        res.recOption.endMargin = 0;
        // ^掞Ԃ
        if (diff <= res.duration) {
            res.duration -= diff;
        }
        else {
            diff -= res.duration;
            res.duration = 0;
            // Jn}[W(ɘ^JnύXKv)
            if (diff > res.recOption.startMargin) diff = res.recOption.startMargin;
            res.recOption.startMargin -= diff;
            res.startTime += -diff * FILETIME_SECOND;
        }
    }

    return true;
}


// ߂̗\폜
bool CReserveList::DeleteNearest(const RECORDING_OPTION &defaultRecOption)
{
    RESERVE *prev;
    RESERVE *pRes = GetNearest(defaultRecOption, &prev);
    if (!pRes) return false;

    CBlockLock lock(&m_writeLock);

    if (!prev) m_head = pRes->next;
    else prev->next = pRes->next;
    delete pRes;
    return true;
}


void CReserveList::SetPluginFileName(LPCTSTR fileName)
{
    CBlockLock lock(&m_writeLock);

    TCHAR saveFileName[MAX_PATH + 32];
    ::lstrcpy(saveFileName, fileName);
    ::PathRemoveExtension(saveFileName);
    ::lstrcat(saveFileName, TEXT("_Reserves"));

    ::lstrcpyn(m_saveTaskName, ::PathFindFileName(saveFileName), ARRAY_SIZE(m_saveTaskName));
    ::lstrcat(saveFileName, TEXT(".txt"));
    ::lstrcpyn(m_saveFileName, saveFileName, ARRAY_SIZE(m_saveFileName));

    ::GetShortPathName(fileName, m_pluginShortPath, ARRAY_SIZE(m_pluginShortPath));
}


bool CReserveList::RunSaveTask(int resumeMargin, int execWait, LPCTSTR tvTestAppName, LPCTSTR driverName,
                               LPCTSTR tvTestCmdOption, HWND hwndPost, UINT uMsgPost)
{
    if (!m_pluginShortPath[0] || !m_saveTaskName[0] || !tvTestAppName[0] || !driverName[0]) return false;

    m_writeLock.Lock();

    // ^XNXPW[o^ɕKvȏZbg
    m_saveTask.resumeMargin = resumeMargin;
    m_saveTask.execWait = execWait;
    ::lstrcpyn(m_saveTask.tvTestAppName, tvTestAppName, ARRAY_SIZE(m_saveTask.tvTestAppName));
    ::lstrcpyn(m_saveTask.driverName, driverName, ARRAY_SIZE(m_saveTask.driverName));
    ::lstrcpyn(m_saveTask.tvTestCmdOption, tvTestCmdOption, ARRAY_SIZE(m_saveTask.tvTestCmdOption));
    m_saveTask.hwndPost = hwndPost;
    m_saveTask.uMsgPost = uMsgPost;

    if (m_hThread) {
        ::WaitForSingleObject(m_hThread, INFINITE);
        ::CloseHandle(m_hThread);
    }
    m_hThread = ::CreateThread(NULL, 0, SaveTaskThread, this, 0, NULL);
    if (!m_hThread) {
        m_writeLock.Unlock();
        return false;
    }

    // Lock̊J̓XbhɔC
    return true;
}


DWORD WINAPI CReserveList::SaveTaskThread(LPVOID pParam)
{
    // oւ݂̏͋֎~!
    CReserveList *pThis = reinterpret_cast<CReserveList*>(pParam);
    CONTEXT_SAVE_TASK *pSaveTask = &pThis->m_saveTask;
    bool fInitialized = false;
    bool fRv = false;
    HRESULT hr;
    ITaskScheduler *pScheduler = NULL;
    ITask *pTask = NULL;
    IPersistFile *pPersistFile = NULL;

    if (FAILED(::CoInitialize(NULL))) goto EXIT;
    fInitialized = true;

    hr = ::CoCreateInstance(CLSID_CTaskScheduler, NULL, CLSCTX_INPROC_SERVER,
                            IID_ITaskScheduler, reinterpret_cast<LPVOID*>(&pScheduler));
    if (hr != S_OK) goto EXIT;

    // resumeMargin<=0̂Ƃ̓^XN폜
    if (pSaveTask->resumeMargin <= 0) {
        pScheduler->Delete(pThis->m_saveTaskName);
        fRv = true;
        goto EXIT;
    }

    hr = pScheduler->Activate(pThis->m_saveTaskName, IID_ITask, reinterpret_cast<IUnknown**>(&pTask));
    if (hr != S_OK) {
        hr = pScheduler->NewWorkItem(pThis->m_saveTaskName, CLSID_CTask, IID_ITask, reinterpret_cast<IUnknown**>(&pTask));
        if (hr != S_OK) {
            pTask = NULL;
            goto EXIT;
        }
    }
    // Rundll32ɃLbN
    TCHAR rundllPath[MAX_PATH];
    if (!GetRundll32Path(rundllPath) || pTask->SetApplicationName(rundllPath) != S_OK) goto EXIT;

    // Rundll32ɓnp[^𐶐
    TCHAR parameters[MAX_PATH * 3 + CMD_OPTION_MAX + 64];
    ::wsprintf(parameters, TEXT("%s,DelayedExecute %d \"%s\" /D \"%s\""), pThis->m_pluginShortPath,
               pSaveTask->execWait, pSaveTask->tvTestAppName, pSaveTask->driverName);
    if (pSaveTask->tvTestCmdOption[0]) {
        ::lstrcat(parameters, TEXT(" "));
        ::lstrcat(parameters, pSaveTask->tvTestCmdOption);
    }
    TCHAR accountName[UNLEN + 1];
    DWORD accountLen = UNLEN + 1;
    if (!::GetUserName(accountName, &accountLen) ||
        pTask->SetParameters(parameters) != S_OK ||
        pTask->SetAccountInformation(accountName, NULL) != S_OK ||
        pTask->SetFlags(TASK_FLAG_RUN_ONLY_IF_LOGGED_ON | TASK_FLAG_SYSTEM_REQUIRED) != S_OK) goto EXIT;

    // ȑÕgKNA
    WORD count = 0;
    for (;;) {
        hr = pTask->GetTriggerCount(&count);
        if (count <= 0) break;
        pTask->DeleteTrigger(0);
    }
    // gKo^
    RESERVE *tail = pThis->m_head;
    for (int i = 0; tail && i < TASK_TRIGGER_MAX; tail = tail->next, i++) {
        FILETIME resumeTime = tail->GetTrimmedStartTime();
        resumeTime += -pSaveTask->resumeMargin * FILETIME_MINUTE;
        SYSTEMTIME resumeSysTime;
        if (!::FileTimeToSystemTime(&resumeTime, &resumeSysTime)) continue;

        TASK_TRIGGER trigger = {0};
        trigger.cbTriggerSize = sizeof(trigger);
        trigger.wBeginYear = resumeSysTime.wYear;
        trigger.wBeginMonth = resumeSysTime.wMonth;
        trigger.wBeginDay = resumeSysTime.wDay;
        trigger.wStartHour = resumeSysTime.wHour;
        trigger.wStartMinute = resumeSysTime.wMinute;
        trigger.TriggerType = TASK_TIME_TRIGGER_ONCE;

        WORD newTrigger;
        ITaskTrigger *pTaskTrigger = NULL;
        if (pTask->CreateTrigger(&newTrigger, &pTaskTrigger) == S_OK) {
            hr = pTaskTrigger->SetTrigger(&trigger);
            pTaskTrigger->Release();
        }
    }
    // ۑ
    hr = pTask->QueryInterface(IID_IPersistFile, reinterpret_cast<void**>(&pPersistFile));
    if (hr != S_OK) goto EXIT;
    hr = pPersistFile->Save(NULL, TRUE);
    if (hr != S_OK) goto EXIT;

    fRv = true;
EXIT:
    if (pSaveTask->hwndPost) {
        ::PostMessage(pSaveTask->hwndPost, pSaveTask->uMsgPost, fRv, 0);
    }
    if (pPersistFile) pPersistFile->Release();
    if (pTask) pTask->Release();
    if (pScheduler) pScheduler->Release();
    if (fInitialized) ::CoUninitialize();
    pThis->m_writeLock.Unlock();
    DEBUG_OUT(TEXT("CReserveList::SaveTaskThread(): "));
    DEBUG_OUT(fRv ? TEXT("SUCCEEDED\n") : TEXT("FAILED\n"));
    return 0;
}


// \ꗗj[쐬
HMENU CReserveList::CreateListMenu(int idStart) const
{
    HMENU hmenu = ::CreateMenu();

    RESERVE *tail = m_head;
    for (int i = 0; tail && i < MENULIST_MAX; i++, tail = tail->next) {
        TCHAR szItem[128];
        SYSTEMTIME sysTime;
        FILETIME time = tail->GetTrimmedStartTime();
        ::FileTimeToSystemTime(&time, &sysTime);
        int len = ::wsprintf(szItem, TEXT("%02hu(%s)%02hu%02hu%s "),
                             sysTime.wDay, GetDayOfWeekText(sysTime.wDayOfWeek),
                             sysTime.wHour, sysTime.wMinute,
                             tail->recOption.IsViewOnly() ? TEXT("") : TEXT(""));
        ::lstrcpyn(szItem + len, tail->eventName, 32);
        // vtBNX΍
        for (LPTSTR p = szItem; *p; ++p) if (*p == TEXT('&')) *p = TEXT('_');
        ::AppendMenu(hmenu, MF_STRING, idStart + i, szItem);
    }
    return hmenu;
}
